// Generated by CoffeeScript 2.4.1
(function() {
  // # Domdom: a simple, dynamic HTML presentation system that supports local or client/server usage

  // Copyright (C) 2019, by Bill Burdick, ZLIB licensed, https://github.com/zot/domdom

  // Domdom uses a JSON object to implement its own Document Object Model that you can share with your local JavaScipt code or with a server. Domdom renders the JSON object in the browser using definitions you provide and it re-renders parts of the GUI when you change values in the JSON object. You can manage the model either in local JavaScript or on a server. Domdom also binds parts of the JSON object and changes it when users interact with the GUI, transmitting those changes to the local JavaScript code or to the server.

  // Domdom is engineered to be simple and lightweight, defined in roughly 600 lines of CoffeeScript.

  // # Overview

  // Domdom chooses a "view" for each nested object in the JSON object you provide by using the object's "type" property. Views are defined using Handlebars, displaying with the JSON object as their context. Domdom also supports namespaces for views, i.e. you can define different views on the same type for different contexts (an object could render as a form in one namespace and a list item in another namespace).

  // When the Javascript model (or server, if connected) changes some of Domdom's JSON objects, it automatically rerenders the views for those objects.

  // Domdom can bind values in its HTML views to paths in its JSON objects so that the HTML can display and/or change the values at thoses paths. When the user changes one of those values, Domdom changes the JSON object at that path and sends an event to the Javascript model (or the server, if connected).

  // # Views

  // Views are Handlebars templates that can also contain other views via the `view` Handlebar plugin and `data-path` attributes in divs and spans which each specify a *path* to property in a JSON object.

  // Input elements in views can also contain `data-path` attributes that specifying a
  // *path* to a property in the JSON object, example:

  // `<input type=text data-path="a.b.c">`

  // If an element has a non-null data-bind-keypress attribute, any keypresses that are not enter or return will be sent as "key" events to the Javascript model (or server, if connected).

  // An element is considered to be a button if it has a data-path property and it is either a non-input element, a button element, or a submit element. The behavior on the JSON object depends on its "value" attribute (if there is one):

  // * no value attribute: when you press the button, Domdom does not change the JSON object but it sends a click event to the model (see Events, below)
  // * the value is a boolean: it acts as a checkbox and when you press it, Domdom sets the boolean value in the JSON object and sends a "set" event (see Events, below)
  // * otherwise: when the input element changes (like by focusing out of a field), Domdom sets the JSON path in the object to the value property, parsed as a JSON value (see Events, below)

  // # Main JSON object

  // ```
  // views: {NAMESPACE: {TYPE: HANDLEBARSDEF}, ...}
  // type: top
  // content: [DATA, ...]
  // ```

  //TODO elaborate on this...
  //The main JSON object supplied to Domdom can optionally provide

  // # Events
  // The Javascript model (or the server, if you are connecting to one) recieves events for clicks and sets with the JSON path and the new value, if there is one. The model (or server) can then change the JSON model in response to trigger an update on the screen, which re-renders the parts of the model that have changed.

  // # Viewdefs

  // You define views with viewdefs and this is normally in the HTML file by putting `data-viewdef` attributes in HTML elements. the value of the `data-viewdef` element can be:

  // - `TYPE`, where TYPE is any string value a JSON object might have in its `type` property
  // - `NAMESPACE/TYPE`, where namespace is any name you choose to make a namespace and TYPE is as above

  // You can use a namespace with the `view` Handlebars plugin (see below).

  // You can also define viewdefs in the `views` property of the main JSON object.

  // Within a viewdef, you can template attributes in two ways see (example-server.html)[../example-server.html]'s account viewdef:

  // 1. enclose the entire contents of the viewdef in an HTML comment
  // 2. put the attribute templating in the value of a data-subst attribute

  // # The namespace type
  // The namespace type sets the namespace for its content object or array of objects, like this:

  // {"type": "namespace", "namespace": "bubba", "content": ...}

  // This will set the namespace to bubba for the content object or array of objects.

  // # The view plugin for Handlebars

  // The predefined `view` plugin lets you show a view on an object or array of objects and you can optionally set the namespace, like this:

  // {{{view `path.in.JSON.object`}}}

  // or

  // {{{view `path.in.JSON.object` `namespace-name`}}}

  // # Events
  // There are two types of events:

  // - set(path, value): the user changed something in the GUI that triggered a set event
  // - click(path, value): the user clicked a button, which can optionally include a value, depending on the view

  // # Controllers
  // If you need custom javascript code, you can use a script element. You an use `element.query()`, `element.queryAll()`, `element.closest()`, etc. to access your view. In addition, these properties will be available:

  // - `document.currentScript` will be the script element (as usual)
  // - `Domdom.currentScript` will also hold the script element
  // - `Domdom.activating` will be true, check this to verify that a view is actually initializing
  // - `Domdom.context` will be the current context object
  // - `Domdom.docPath` will be the current docPath (see DocPaths, below)

  // Also, each view will have a `data-location` attribute set to its path and a `data-namespace` attribute set to the view's namespace.

  // # Using Domdom

  // On the web side, you need to make sure the files in the js and css directories are available to your HTML file and include these elements (altered to fit your file layout, of course):

  // \<link rel="stylesheet" href="css/domdom.css">\</link>
  // \<script src="js/lib/handlebars-v4.0.5.js">\</script>
  // \<script src="js/domdom.js">\</script>

  // It's also compatible with AMD style so you can use something like require.js:

  // \<link rel="stylesheet" href="css/domdom.css">\</link>
  // \<script data-main="js/config" src="js/lib/require-2.1.18.js">\</script>

  // You can implement the model in local JavaScript or in a server. Domdom currently supports Julia servers.

  // # Connecting to a server
  // Put this at the bottom of the body of your web page, with the HOST and PORT of your server in it:

  // \<script>Domdom.connect({}, "ws://HOST:PORT")\</script>

  // The Julia server code supports its own version of event handlers and DocPath (see the JavaScript model documentation below)

  // # Using Domdom with a JavaScript model

  // * Create a Javascript object with
  // ```
  // {type: 'document',
  //  views: {default: {viewdefs...},
  //  NAMESPACE1: {viewdefs...}},
  //  contents: [CONTENTS...]}
  // ```

  // Views are optional in the object since they can also be in the HTML page.

  // - Create a context with {top: JSON, handler: HANDLER}
  //   - JSON is the JSON object you have created
  //   - HANDLER is an event handler
  //     - You can use the patternHandler() function to easily specify event handlers (see source for documentation).
  //     - Otherwise, the handler is {clickButton: (evt)=> ..., changedValue: (evt)=> ..., key: (evt)=> ...}
  //     - the dispatchClick, dispatchKey, and dispatchSet functions dispatch events in a high-level way, using DocPaths (see below)
  // ## DocPaths
  // A DocPath is proxy that makes it easy navigate paths in the JSON object and it lets you change the JSON object and automatically trigger re-rendering for those changes. It's called DocPath because the JSON object is the "document" of the Document Object Model. PatternHandler and the three dispatch functions (dispatchClick, dispatchKey, and dispatchSet) each send a DocPath as the first argument to your provided event handler function.

  // Given docp is a DocPath...

  // - `docp.PROP` returns the value in the document at PROP if it is atomic or, if the value is an array or object, it returns a new DocPath for that location (with docp's path extended to include PROP)
  // - `docp[INDEX]` returns the value in the document at INDEX if it is atomic or, if the value is an array or object, it returns a new DocPath for that location (with docp's path extended to include INDEX)
  // - `docPathValue(docp)` returns docp's value
  // - `docp.PROP = VALUE` sets the value in the document and cause Domdom to re-render it
  // - `docPathParts(docp)` returns the "parts" of a DocPath, the Domdom object, the context, and the path array

  // You can use `batch(con, func)` if you need to change DocPaths outside of an event handler for "event compression". Batch eliminates re-rendering of the same object multiple times.

  // # History

  // I came up with the original concept around 2000 or 2001, as the next step in evolution for Classic Blend (a remote presentation system I first developed in 1995). The idea of the next step was that if you abstracted an entire GUI into a set of shared variables, you could use the variables to control a remote GUI from a server kind of like a [tuple space](https://en.wikipedia.org/wiki/Tuple_space) or like [SNMP](https://en.wikipedia.org/wiki/Simple_Network_Management_Protocol). Beyond this, you could reskin the GUI in dramatically different ways -- far more radically than GTK themes, for instance -- switching from a web browser to the Unreal engine, for example, where menus might be presented as shops (I actually prototyped a Quake-based front end at one point).

  // I've been using an earlier and quite different variation of this idea since 2006 on an extremely large project. The browser side of the presentation is fully automatic now and we don't write any JavaScript for our front ends anymore, unless we're adding new kinds of widgets.

  // This version of the concept, Domdom, grew out of the Leisure project (which will eventually be updated to use Domdom) and I've used variations of this JavaScript and server code in several of my personal projects.

  // The [Xus](https://github.com/zot/Xus) project is also related to this and it's also based on shared variables.
  var define, ref,
    indexOf = [].indexOf;

  define = (ref = window.define) != null ? ref : function(n, func) {
    return window.Domdom = func(window.Handlebars);
  };

  define(['handlebars'], function(Handlebars) {
    var Domdom, _verbose, batch, change, compile, connect, curId, dispatchClick, dispatchKey, dispatchSet, docPath, docPathParent, docPathParts, docPathValue, docValue, domdomBlur, domdomFocus, domdoms, eventPaths, find, findIds, globalContext, handleChanges, handleMessage, initChangeContext, isDocPath, isDocPathSym, isRefreshing, keyCode, locationFor, locationToString, messages, normalizePath, parseHtml, parsingDiv, partsSym, patternHandler, query, queryAll, registerHelper, replace, resolvePath, setVerbose, stringToLocation, syms, takeFocusForScript, verbose;
    ({compile, registerHelper} = Handlebars);
    curId = 0;
    _verbose = false;
    setVerbose = function(value) {
      return _verbose = value;
    };
    verbose = function(...args) {
      return _verbose && console.log(...args);
    };
    keyCode = function(evt) {
      var key, ref1;
      if (!((ref1 = evt.key.toLowerCase()) === 'shift' || ref1 === 'control' || ref1 === 'alt')) {
        key = evt.key;
        if (key.toLowerCase().startsWith('arrow')) {
          key = key.slice(5).toLowerCase();
        }
        if (evt.shiftKey && key.length > 1) {
          key = "S-" + key;
        }
        if (evt.ctrlKey) {
          key = "C-" + key;
        }
        if (evt.altKey || evt.metaKey) {
          key = "M-" + key;
        }
        return key;
      }
    };
    parsingDiv = document.createElement('div');
    query = document.querySelector.bind(document);
    queryAll = document.querySelectorAll.bind(document);
    find = function(node, selector, includeSelf) {
      if (includeSelf && node.matches(selector)) {
        return [node].concat(Array.prototype.slice.call(node.querySelectorAll(selector)));
      } else {
        return node.querySelectorAll(selector);
      }
    };
    parseHtml = function(str) {
      var dom;
      parsingDiv.innerHTML = `<div>${str}</div>`;
      dom = parsingDiv.firstChild;
      parsingDiv.innerHTML = '';
      if (dom.childNodes.length === 1 && dom.firstChild.nodeType === 1) {
        return dom.firstChild;
      } else {
        return dom;
      }
    };
    locationToString = function(loc) {
      var coord, l, len, str;
      str = "";
      for (l = 0, len = loc.length; l < len; l++) {
        coord = loc[l];
        if (str) {
          str += " ";
        }
        str += coord;
      }
      return str;
    };
    stringToLocation = function(str) {
      var coord, l, len, ref1, results;
      if (str === "") {
        return [];
      }
      ref1 = str.split(' ');
      results = [];
      for (l = 0, len = ref1.length; l < len; l++) {
        coord = ref1[l];
        results.push(String(Number(coord)) === coord ? Number(coord) : coord);
      }
      return results;
    };
    locationFor = function(json, context) {
      var ref1;
      return ((ref1 = context.top.index[json.$ID$]) != null ? ref1[1] : void 0) || context.location;
    };
    resolvePath = function(doc, location) {
      var first, j, parent, path;
      if (typeof location === 'string') {
        [j, path, parent] = doc.index[location];
        location = path;
      }
      if (typeof location[0] === 'string' && location[0][0] === '@') {
        first = location[0].slice(1);
        [j, path, parent] = doc.index[first];
        location = location.length > 1 ? [...path, ...location.slice(1)] : path;
      }
      return location;
    };
    normalizePath = function(path, index) {
      var ignore;
      if (Array.isArray(path)) {
        return path;
      } else {
        [ignore, path] = index[typeof path === 'object' ? path.$ID$ : path];
        return path;
      }
    };
    findIds = function(parent, json, location = [], items = {}) {
      var el, i, k, l, len, loc, v;
      if (Array.isArray(json)) {
        for (i = l = 0, len = json.length; l < len; i = ++l) {
          el = json[i];
          findIds(json, el, [...location, i], items);
        }
      } else if (json !== null && typeof json === 'object' && (json.type != null)) {
        loc = locationToString(location);
        items[loc] = [json, location];
        for (k in json) {
          v = json[k];
          findIds(json, v, [...location, k], items);
        }
      }
      return items;
    };
    globalContext = {
      namespace: 'default'
    };
    replace = function(oldDom, newDom) {
      var l, len, n, na, nav, ref1;
      // prefer mutating the old dom to replacing it
      if (oldDom && oldDom.nodeName === newDom.nodeName && oldDom.childNodes.length === 0 && newDom.childNodes.length === 0) {
        na = new Set(newDom.getAttributeNames());
        ref1 = oldDom.getAttributeNames();
        for (l = 0, len = ref1.length; l < len; l++) {
          n = ref1[l];
          if (!na.has(n)) {
            oldDom.removeAttribute(n);
          }
        }
        for (n of na) {
          nav = newDom.getAttribute(n);
          if (nav !== oldDom.getAttribute(n)) {
            oldDom.setAttribute(n, nav);
          }
        }
        return oldDom;
      } else {
        oldDom.replaceWith(newDom);
        return newDom;
      }
    };
    domdoms = [];
    domdomBlur = function(event) {
      var l, len, md, results;
      results = [];
      for (l = 0, len = domdoms.length; l < len; l++) {
        md = domdoms[l];
        if (event.target.nodeType === 1 && md.top.contains(event.target)) {
          results.push(md.blurring = true);
        } else {
          results.push(void 0);
        }
      }
      return results;
    };
    domdomFocus = function(event) {
      var l, len, md, results;
      results = [];
      for (l = 0, len = domdoms.length; l < len; l++) {
        md = domdoms[l];
        if (md.blurring) {
          md.blurring = false;
          results.push(md.runRefreshQueue());
        } else {
          results.push(void 0);
        }
      }
      return results;
    };
    takeFocusForScript = function(sel) {
      var el;
      if (isRefreshing(Domdom.currentScript)) {
        el = find(Domdom.currentScript.parentElement, sel)[0];
        return setTimeout((function() {
          console.log(document.activeElement);
          return el.focus();
        }), 1);
      }
    };
    isRefreshing = function(node) {
      var p, path, ref1, ref2;
      if (Domdom.activating && Domdom.refreshLocations) {
        path = (ref1 = node.closest('[data-location]')) != null ? ref1.getAttribute('data-location') : void 0;
        if (Domdom.refreshLocations.size === 0) {
          return true;
        }
        ref2 = Domdom.refreshLocations;
        for (p of ref2) {
          if (path.startsWith(p)) {
            return true;
          }
        }
      }
    };
    Domdom = class Domdom {
      constructor(top1) {
        this.top = top1;
        if (!this.top) {
          throw new Error("No top node for Domdom");
        }
        this.refreshQueue = []; // queued refresh commands that execute after the current event
        this.refreshLocations = new Set();
        this.specialTypes = {
          document: (dom, json, context) => {
            return this.renderTop(dom, json, context);
          }
        };
        if (!domdoms.length) {
          window.addEventListener("blur", domdomBlur, true);
          window.addEventListener("focus", domdomFocus, true);
        }
        domdoms.push(this);
      }

      // activateScripts inserts copies of the parsed script elements, which makes them execute.
      activateScripts(el, ctx) {
        var l, len, newScript, ref1, results, script, text;
        if (!Domdom.activating) {
          Domdom.activating = true;
          Domdom.context = ctx;
          Domdom.docPath = docPath(this, ctx, ctx.location);
          try {
            ref1 = el.querySelectorAll('script');
            results = [];
            for (l = 0, len = ref1.length; l < len; l++) {
              script = ref1[l];
              if ((!script.type || script.type.toLowerCase() === 'text/javascript') && (text = script.textContent)) {
                newScript = document.createElement('script');
                newScript.type = 'text/javascript';
                if (script.src) {
                  newScript.src = script.src;
                }
                newScript.textContent = text;
                //keep the current script here in case the code needs access to it
                Domdom.currentScript = newScript;
                script.parentNode.insertBefore(newScript, script);
                results.push(script.parentNode.removeChild(script));
              } else {
                results.push(void 0);
              }
            }
            return results;
          } finally {
            Domdom.currentScript = null;
            Domdom.activating = false;
            Domdom.context = null;
          }
        }
      }

      // Find view for json and replace dom with the rendered view. Context contains global info like the
      // current namespace, etc.
      render(dom, json, context) {
        var newDom;
        if (context.views == null) {
          context.views = {};
        }
        newDom = this.baseRender(dom, json, Object.assign({
          location: []
        }, context));
        this.analyzeInputs(newDom, context, Array.isArray(json));
        return newDom;
      }

      baseRender(dom, json, context) {
        var childDom, el, i, id, l, len, newDom, ref1, ref2, special;
        context = Object.assign({}, globalContext, context);
        if (Array.isArray(json)) {
          newDom = dom;
          for (i = l = 0, len = json.length; l < len; i = ++l) {
            childDom = json[i];
            el = document.createElement('div');
            newDom.appendChild(el);
            this.baseRender(el, childDom, Object.assign({}, context, {
              location: [...context.location, i]
            }));
          }
          return newDom;
        } else {
          id = (ref1 = (ref2 = json.$ID$) != null ? ref2 : dom.getAttribute('id')) != null ? ref1 : ++curId;
          dom.setAttribute('id', id);
          if (special = this.specialTypes[json.type]) {
            return special(dom, json, context);
          } else {
            return this.normalRender(dom, json, context);
          }
        }
      }

      // special renderers can use this to modify how their views render
      normalRender(dom, json, context) {
        var def, newDom, old;
        def = this.findViewdef(json.type, context);
        newDom = parseHtml((function() {
          if (def) {
            try {
              old = globalContext;
              globalContext = context;
              return def(json, {
                data: Object.assign({
                  domdom: this
                }, {context})
              });
            } finally {
              globalContext = old;
            }
          } else {
            return `COULD NOT RENDER TYPE ${json.type}, NAMESPACE ${context.namespace}`;
          }
        }).call(this));
        newDom.setAttribute('data-location', locationToString(context.location));
        if (!newDom.getAttribute('data-namespace')) {
          newDom.setAttribute('data-namespace', context.namespace);
        }
        newDom.setAttribute('id', json.$ID$);
        if (newDom.getAttribute('data-path')) {
          newDom.setAttribute('data-path-full', locationToString(context.location));
        }
        newDom = replace(dom, newDom);
        this.populateInputs(newDom, json, context);
        this.activateScripts(newDom, context);
        return newDom;
      }

      findViewdef(type, context) {
        var def, domClone, el, namespace, r1, r2, ref1, ref2, ref3, ref4;
        if (def = (ref1 = context.views) != null ? (ref2 = ref1[context.namespace]) != null ? ref2[type] : void 0 : void 0) {
          return def;
        } else if (el = query(`[data-viewdef='${type}/${context.namespace}']`)) {
          namespace = context.namespace;
        } else if (def = (ref3 = context.views) != null ? (ref4 = ref3.default) != null ? ref4[type] : void 0 : void 0) {
          return def;
        } else if (!(el = query(`[data-viewdef='${type}']`))) {
          return null;
        }
        if (context.views == null) {
          context.views = {};
        }
        if (context.views[namespace] == null) {
          context.views[namespace] = {};
        }
        domClone = el.cloneNode(true);
        domClone.removeAttribute('data-viewdef');
        if (domClone.firstChild && domClone.firstChild.nodeType === Node.COMMENT_NODE && domClone.firstChild === domClone.lastChild) {
          return context.views[namespace][type] = compile(domClone.outerHTML.replace(/<!--((.|\n)*)-->/, '$1'));
        } else {
          r1 = /data-subst=\'([^\']*)\'/;
          r2 = /data-subst=\"([^\"]*)\"/;
          return context.views[namespace][type] = compile(domClone.outerHTML.replace(r1, '$1').replace(r2, '$1'));
        }
      }

      rerender(json, context, thenBlock, exceptNode) {
        this.refreshLocations.add(locationToString(context.location));
        return this.queueRefresh(() => {
          var l, len, len1, m, newDom, node, oldDom, oldLocation, ref1, ref2, top;
          ref1 = this.domsForRerender(json, context);
          for (l = 0, len = ref1.length; l < len; l++) {
            oldDom = ref1[l];
            if (!(!exceptNode || oldDom !== exceptNode)) {
              continue;
            }
            oldLocation = oldDom.getAttribute('data-location');
            context = Object.assign({}, context, {
              rerender: locationFor(json, context)
            });
            if (oldLocation === locationToString(locationFor(json, context)) && oldDom.getAttribute('data-namespace')) {
              context.namespace = oldDom.getAttribute('data-namespace');
            }
            newDom = context.location.length === 1 ? this.renderTop(oldDom, context.top, context) : newDom = this.render(oldDom, json, context);
            top = newDom.closest('[data-top]');
            ref2 = find(newDom, '[data-path-full]');
            for (m = 0, len1 = ref2.length; m < len1; m++) {
              node = ref2[m];
              this.valueChanged(top, node, context);
            }
          }
          return thenBlock(newDom);
        });
      }

      // domsForRender(json, context) finds the doms for json or creates and inserts a blank one
      domsForRerender(json, context) {
        var dom, end, inside, l, len, len1, loc, location, m, node, parent, parentDom, ref1, targets;
        if (!json) {
          ref1 = queryAll(`[data-location^='${locationToString(context.location)}']`);
          for (l = 0, len = ref1.length; l < len; l++) {
            node = ref1[l];
            node.remove();
          }
          return [];
        }
        if (location = locationFor(json, context)) {
          loc = locationToString(context.location);
          targets = queryAll(`[data-location^='${loc}']`);
          if (targets.length === 1) {
            return targets;
          }
          for (m = 0, len1 = targets.length; m < len1; m++) {
            node = targets[m];
            if (node.getAttribute('data-location') !== loc && node.parentNode.closest(`[data-location^='${loc}']`)) {
              node.remove();
            }
          }
          if (targets.length) {
            return targets;
          }
          inside = false;
          end = location[location.length - 1];
          if (typeof end === 'number') {
            parent = this.getPath(context.top, context.top.contents, [...location.slice(0, -1), end - 1]);
          } else {
            parent = this.getPath(context.top, context.top.contents, location.slice(0, -1));
            inside = true;
          }
          if (parentDom = query(`[id='${parent.$ID$}']`)) {
            dom = parseHtml(`<div id='${json.$ID$}' data-location='${locationToString(location)}'></div>`);
            if (inside) {
              parentDom.appendChild(dom);
            } else {
              parentDom.after(dom);
            }
          }
          return [dom];
        }
        return [];
      }

      renderTop(dom, json, context) {
        var base, base1, contents, def, k, namespace, newDom, ref1, type, types, v, views;
        ({views, contents} = json);
        json.index = {};
        ref1 = findIds(null, contents);
        for (k in ref1) {
          v = ref1[k];
          if (!v.$ID$) {
            v[0].$ID$ = ++curId;
          }
          json.index[v[0].$ID$] = v;
        }
        json.compiledViews = {};
        if (context.views == null) {
          context.views = {};
        }
        for (namespace in views) {
          types = views[namespace];
          if ((base = json.compiledViews)[namespace] == null) {
            base[namespace] = {};
          }
          if ((base1 = context.views)[namespace] == null) {
            base1[namespace] = {};
          }
          for (type in types) {
            def = types[type];
            //destructively modify context's views
            context.views[namespace][type] = json.compiledViews[namespace][type] = compile(def);
          }
        }
        newDom = this.baseRender(dom, contents.main, Object.assign(context, {
          top: json,
          location: ['main']
        }));
        newDom.setAttribute('data-top', 'true');
        return newDom;
      }

      queueRefresh(cmd) {
        this.refreshQueue.push(cmd);
        if (!this.pressed && !this.blurring && document.activeElement !== document.body) {
          return this.runRefreshQueue();
        }
      }

      runRefreshQueue() {
        var q;
        if (this.refreshQueue.length > 0) {
          q = this.refreshQueue;
          this.refreshQueue = [];
          return setTimeout((() => {
            var active, cmd, field, focusPath, l, len;
            Domdom.refreshLocations = this.refreshLocations;
            active = document.activeElement;
            focusPath = active && active.getAttribute('data-path-full');
            for (l = 0, len = q.length; l < len; l++) {
              cmd = q[l];
              cmd();
            }
            this.refreshLocations.clear();
            Domdom.refreshLocations = null;
            if (focusPath) {
              field = query(`[data-path-full='${focusPath}']`);
              if (field != null) {
                field.focus();
              }
              return field != null ? typeof field.select === "function" ? field.select() : void 0 : void 0;
            }
          }), 5);
        }
      }

      adjustIndex(index, path, parent, oldJson, newJson) {
        var k, newIds, oldIds, oldKeys, ref1, results, v;
        oldIds = findIds(parent, oldJson, path);
        newIds = findIds(parent, newJson, path);
        oldKeys = new Set(Object.keys(oldIds));
        for (k in newIds) {
          v = newIds[k];
          if (!v[0].$ID$ && ((ref1 = oldIds[k]) != null ? ref1[0].$ID$ : void 0)) {
            v[0].$ID$ = oldIds[k][0].$ID$;
          } else if (!v[0].$ID$) {
            v[0].$ID$ = ++curId;
          }
          index[v[0].$ID$] = v;
          oldKeys.delete(k);
        }
        results = [];
        for (k of oldKeys) {
          results.push(delete index[oldIds[k][0].$ID$]);
        }
        return results;
      }

      analyzeInputs(dom, context, array) {
        var l, len, node, ref1, results;
        ref1 = find(dom, "input, textarea, button, [data-path]", true);
        results = [];
        for (l = 0, len = ref1.length; l < len; l++) {
          node = ref1[l];
          if (!array || node !== dom) {
            results.push(((node) => {
              var attr, fullpath, len1, m, newDom, path, ref2, ref3, ref4, ref5, renderingNode, results1, subcontext;
              if (fullpath = node.getAttribute('data-path-full')) {
                path = stringToLocation(node.getAttribute('data-path-full'));
                if (node.getAttribute('data-bind-keypress')) {
                  node.on('keydown', function(e) {
                    var base, ref2;
                    if (!((ref2 = keyCode(e)) === 'C-r' || ref2 === 'C-J')) {
                      e.preventDefault();
                      e.stopPropagation();
                      return typeof (base = context.handler).keyPress === "function" ? base.keyPress(e.originalEvent) : void 0;
                    }
                  });
                }
                if ((ref2 = node.nodeName) === 'DIV' || ref2 === 'SPAN') { // handle data-path in divs and spans
                  if (node.getAttribute('data-replace')) {
                    node.innerHTML = '';
                    renderingNode = node;
                  } else {
                    node.innerHTML = '<div></div>';
                    renderingNode = node.firstChild;
                  }
                  path = stringToLocation(node.getAttribute('data-path-full'));
                  subcontext = Object.assign({}, context, {
                    location: path,
                    namespace: node.getAttribute('data-namespace')
                  });
                  newDom = this.render(renderingNode, this.getPath(context.top, context.top.contents, path), subcontext);
                  node.removeAttribute('data-path');
                  ref3 = ['style', 'class'];
                  results1 = [];
                  for (m = 0, len1 = ref3.length; m < len1; m++) {
                    attr = ref3[m];
                    if (node.getAttribute(attr) && !newDom.getAttribute(attr)) {
                      results1.push(newDom.setAttribute(attr, node.getAttribute(attr)));
                    } else {
                      results1.push(void 0);
                    }
                  }
                  return results1;
                } else if (((ref4 = node.type) === 'button' || ref4 === 'submit') || !((ref5 = node.type) === 'text' || ref5 === 'password')) {
                  // using onmousedown, onclick, path, and @pressed because
                  // the view can render out from under the button if focus changes
                  // which replaces the button with ta new one in the middle of a click event
                  node.onmousedown = (evt) => {
                    return this.pressed = path;
                  };
                  return node.onclick = (evt) => {
                    var base, err, newValue, oldValue, v;
                    if (this.pressed === path || evt.detail === 0) {
                      this.pressed = false;
                      newValue = (function() {
                        if (v = node.getAttribute('value')) {
                          try {
                            return JSON.parse(v);
                          } catch (error) {
                            err = error;
                            return v;
                          }
                        } else if ((typeof (oldValue = this.getPath(context.top, context.top.contents, path))) === 'boolean') {
                          return newValue = !oldValue;
                        }
                      }).call(this);
                      if (newValue) {
                        this.setValueFromUser(node, evt, dom, context, path, newValue);
                      } else {
                        if (typeof (base = context.handler).clickButton === "function") {
                          base.clickButton(evt);
                        }
                      }
                      return this.runRefreshQueue();
                    }
                  };
                } else {
                  return node.onchange = (evt) => {
                    var ownerPath, ownerPathString;
                    ownerPathString = evt.srcElement.closest('[data-location]').getAttribute('data-location');
                    ownerPath = stringToLocation(ownerPathString);
                    return this.setValueFromUser(node, evt, dom, context, path, node.value);
                  };
                }
              }
            })(node));
          }
        }
        return results;
      }

      setValueFromUser(node, evt, dom, context, path, value) {
        var base, json, ownerPath, ownerPathString;
        ownerPathString = node.closest('[data-location]').getAttribute('data-location');
        ownerPath = stringToLocation(ownerPathString);
        json = this.getPath(context.top, context.top.contents, ownerPath);
        this.setPath(context.top, context.top.contents, path, value);
        return typeof (base = context.handler).changedValue === "function" ? base.changedValue(evt, value) : void 0;
      }

      populateInputs(dom, json, context) {
        var fullpath, l, len, location, node, path, ref1, ref2, results, setSome;
        ref1 = find(dom, "[data-path]", true);
        results = [];
        for (l = 0, len = ref1.length; l < len; l++) {
          node = ref1[l];
          location = stringToLocation(node.closest('[data-location]').getAttribute('data-location'));
          path = node.getAttribute('data-path').split('.');
          fullpath = locationToString([...location, ...path]);
          if ((ref2 = node.type) === 'text' || ref2 === 'password') {
            node.setAttribute('value', this.getPath(context.top, json, path));
          }
          node.setAttribute('data-path-full', fullpath);
          setSome = true;
          results.push(setSome);
        }
        return results;
      }

      valueChanged(dom, source, context) {
        var attr, fullpath, json, l, len, len1, m, node, ref1, ref2, remove, results, setAttr, value;
        value = source.value;
        fullpath = source.getAttribute('data-path-full');
        json = this.getPath(context.top, context.top.contents, (stringToLocation(fullpath)).slice(0, -1));
        ref1 = find(dom, `[data-path-full='${fullpath}']`);
        results = [];
        for (l = 0, len = ref1.length; l < len; l++) {
          node = ref1[l];
          if (!(node !== source)) {
            continue;
          }
          node.value = value;
          remove = new Set();
          ref2 = node.attributes;
          for (m = 0, len1 = ref2.length; m < len1; m++) {
            attr = ref2[m];
            if (attr.name.startsWith('data-attribute-')) {
              setAttr = attr.name.substring('data-attribute-'.length);
              if (json[attr.value] != null) {
                attr[setAttr] = json[attr.value];
              } else if (node.hasAttribute(setAttr)) {
                remove.add(setAttr);
              }
            }
          }
          results.push((function() {
            var results1;
            results1 = [];
            for (attr of remove) {
              results1.push(node.removeAttribute(attr));
            }
            return results1;
          })());
        }
        return results;
      }

      getPath(doc, json, location) {
        var i, l, len;
        location = resolvePath(doc, location);
        for (l = 0, len = location.length; l < len; l++) {
          i = location[l];
          json = json[i];
        }
        return json;
      }

      setPath(document, json, location, value) {
        var i, index, l, last, lastI, len, newJson, results;
        last = json;
        lastI = 0;
        location = resolvePath(document, location);
        results = [];
        for (index = l = 0, len = location.length; l < len; index = ++l) {
          i = location[index];
          if (index + 1 < location.length) { //not at the end
            last = json;
            lastI = i;
            results.push(json = json[i]);
          } else {
            if ((value != null ? value.type : void 0) != null) {
              this.adjustIndex(document.index, location.slice(0, +index + 1 || 9e9), json, json[i], value);
            } else {
              newJson = Object.assign({}, json);
              newJson[i] = value;
              this.adjustIndex(document.index, location.slice(0, index), last, json, newJson);
            }
            results.push(json[i] = value);
          }
        }
        return results;
      }

      defView(context, namespace, type, def) {
        return context.views[namespace][type] = compile(def);
      }

    };
    Handlebars.registerHelper('view', function(itemName, namespace, options) {
      var context, item, node;
      if (typeof itemName !== 'string' || (namespace && options && typeof namespace !== 'string')) {
        throw new Error("View must be called with one or two strings");
      }
      if (options == null) {
        options = namespace;
        namespace = null;
      }
      item = this[itemName];
      context = options.data.context;
      context = Object.assign({}, context, {
        location: [...context.location, itemName]
      });
      if (namespace) {
        context.namespace = namespace;
      }
      node = options.data.domdom.baseRender(parseHtml('<div></div>'), item, context);
      if (node.nodeType === 1) {
        return node.outerHTML;
      } else {
        return node.data;
      }
    });
    Handlebars.registerHelper('ref', function(item, namespace, options) {
      var context, json, location, node;
      if (!Array.isArray(item) && typeof item !== 'string') {
        throw new Error("Ref must be called with an array or a string and optionally another string");
      }
      if ((options != null) && namespace && typeof namespace !== 'string') {
        throw new Error("Ref's namespace  must be a string");
      }
      if (options == null) {
        options = namespace;
        namespace = null;
      }
      context = options.data.context;
      location = resolvePath(context.top, (typeof item === 'string' ? stringToLocation(item) : item));
      context = Object.assign({}, context, {
        location: location
      });
      if (namespace) {
        context.namespace = namespace;
      }
      if (json = options.data.domdom.getPath(context.top, context.top.contents, location)) {
        node = options.data.domdom.baseRender(parseHtml('<div></div>'), json, context);
        if (node.nodeType === 1) {
          return node.outerHTML;
        } else {
          return node.data;
        }
      } else {
        return "";
      }
    });
    // Command processor clients (if using client/server)
    messages = {
      batch: function(con, items) {
        var item, l, len;
        con.batchLevel++;
        for (l = 0, len = items.length; l < len; l++) {
          item = items[l];
          handleMessage(con, item);
        }
        return con.batchLevel--;
      },
      document: function(con, doc) {
        con.document = doc;
        con.context.top = doc;
        con.dom = con.dd.render(con.dom, doc, con.context);
        con.context.views = con.document.compiledViews;
        return console.log("document:", doc);
      },
      set: function(con, path, value) {
        con.dd.setPath(con.document, con.document.contents, path, value);
        if ((value != null ? value.type : void 0) == null) {
          path.pop();
        }
        con.changedJson.add(locationToString(path));
        return path;
      },
      splice: function(con, path, start, length, ...items) {
        var i, l, obj, p, ref1, ref2;
        obj = con.dd.getPath(con.document, con.document.contents, path);
        for (i = l = ref1 = start, ref2 = obj.length; (ref1 <= ref2 ? l < ref2 : l > ref2); i = ref1 <= ref2 ? ++l : --l) {
          p = [...path, i];
          con.dd.adjustIndex(con.document.index, p, obj, obj[i], items[i - start]);
          con.changedJson.add(locationToString(p));
        }
        return obj.splice(start, length, ...items);
      },
      defView: function(con, namespace, type, def) {
        return con.dd.defView(con.context, namespace, type, def);
      }
    };
    // #Change handler
    handleChanges = function(ctx) {
      var loc, path, ref1;
      if (ctx.batchLevel === 0) {
        ref1 = ctx.changedJson;
        for (path of ref1) {
          loc = stringToLocation(path);
          ctx.dd.rerender(ctx.dd.getPath(ctx.doc, ctx.doc.contents, loc), Object.assign({}, ctx, {
            location: loc
          }), function(dom) {
            if (dom.getAttribute('data-top') != null) {
              return ctx.setTopFunc(dom);
            }
          });
        }
        ctx.changedJson.clear();
        return ctx.dd.runRefreshQueue();
      }
    };
    change = function(ctx, path) {
      ctx.changedJson.add(locationToString(path));
      if (ctx.batchLevel === 0) {
        return handleChanges(ctx);
      }
    };
    initChangeContext = function(dd, ctx, doc, setTopFunc) {
      ctx.batchLevel = 0;
      ctx.changedJson = new Set();
      ctx.dd = dd;
      ctx.doc = doc;
      return ctx.setTopFunc = setTopFunc;
    };
    // `batch(CTX, FUNC)` executes FUNC, queuing up re-rendering requests and then processing the requests all at once after FUNC finishes.
    batch = function(ctx, func) {
      if (typeof ctx.batchLevel === 'number') {
        ctx.batchLevel++;
        try {
          return func();
        } finally {
          ctx.batchLevel--;
          handleChanges(ctx);
        }
      } else {
        return func();
      }
    };
    // #Local Code
    isDocPathSym = Symbol("isDocPath");
    partsSym = Symbol("parts");
    syms = [isDocPathSym, partsSym];
    // `isDocPath(obj)` returns true if the object is a docPath
    isDocPath = function(obj) {
      return obj[isDocPathSym];
    };
    // `docPathParts(docp)` returns the "parts" you used to create the doc path: [dd, ctx, path]
    docPathParts = function(docp) {
      return docp[partsSym];
    };
    // `docPathParent(docp)` returns a parent DocPath for docp (i.e. a DocPath without the last path element)
    docPathParent = function(docp) {
      var ctx, dd, path;
      [dd, ctx, path] = docPathParts(docp);
      if (!path.length) {
        return docp;
      } else {
        return docPath(dd, ctx, path.slice(0, -1));
      }
    };
    // `docPathValue(docp)` returns the value for the DocPath
    docPathValue = function(docp) {
      var ctx, dd, path;
      [dd, ctx, path] = docPathParts(docp);
      return docValue(dd, ctx, path);
    };
    // `docPath(dd, ctx, path = [])` creates a DocPath
    docPath = function(dd, ctx, path = []) {
      return new Proxy({}, {
        get: function(target, name) {
          if (name === isDocPathSym) {
            return true;
          } else if (name === partsSym) {
            return [dd, ctx, path];
          } else if (name === 'toString()') {
            return function() {
              return printDocPath(dd, ctx, path);
            };
          } else {
            return docValue(dd, ctx, [...path, name]);
          }
        },
        set: function(target, name, value) {
          if (!(indexOf.call(syms, name) >= 0)) {
            path = [...path, name];
            dd.setPath(ctx.top, ctx.top.contents, path, value);
            return change(ctx, resolvePath(ctx.top, path));
          }
        }
      });
    };
    docValue = function(dd, ctx, path, value) {
      var val;
      if (value === void 0) {
        val = dd.getPath(ctx.top, ctx.top.contents, path);
        if (val === void 0) {
          return val;
        } else {
          return docValue(dd, ctx, path, dd.getPath(ctx.top, ctx.top.contents, path));
        }
      } else if (Array.isArray(value)) {
        return docPath(dd, ctx, path);
      } else if (typeof value === 'object') {
        return docPath(dd, ctx, path);
      } else {
        return value;
      }
    };
    eventPaths = function(dd, ctx, evt) {
      var fullPath, node, obj, objPath, path, value;
      node = evt.srcElement;
      fullPath = stringToLocation(node.getAttribute('data-path-full'));
      path = stringToLocation(node.getAttribute('data-path'));
      objPath = stringToLocation(node.closest('[data-location]').getAttribute('data-location'));
      obj = dd.getPath(ctx.top, ctx.top.contents, objPath);
      value = dd.getPath(ctx.top, ctx.top.contents, [...objPath, ...path]);
      return [fullPath, obj, objPath, path, value];
    };
    dispatchClick = function(dd, ctx, handlers, evt) {
      var docp, fullPath, obj, objPath, path, value;
      [fullPath, obj, objPath, path, value] = eventPaths(dd, ctx, evt);
      docp = docPath(dd, ctx, fullPath);
      return batch(ctx, function() {
        var name1, ref1, ref2;
        if (typeof handlers[name1 = [obj.type, locationToString(path), "click"].join(',')] === "function") {
          handlers[name1](docp, obj, objPath, path, value, evt);
        }
        return (ref1 = handlers[obj.type]) != null ? (ref2 = ref1[locationToString(path)]) != null ? typeof ref2.click === "function" ? ref2.click(docp, obj, objPath, path, value, evt) : void 0 : void 0 : void 0;
      });
    };
    dispatchKey = function(dd, ctx, handlers, evt) {
      var docp, fullPath, obj, objPath, path, value;
      [fullPath, obj, objPath, path, value] = eventPaths(dd, ctx, evt);
      docp = docPath(dd, ctx, fullPath);
      return batch(ctx, function() {
        var name1, ref1, ref2;
        if (typeof handlers[name1 = [obj.type, locationToString(path), "key"].join(',')] === "function") {
          handlers[name1](docp, obj, objPath, path, value, evt);
        }
        return (ref1 = handlers[obj.type]) != null ? (ref2 = ref1[locationToString(path)]) != null ? typeof ref2.key === "function" ? ref2.key(docp, obj, objPath, path, value, evt) : void 0 : void 0 : void 0;
      });
    };
    dispatchSet = function(dd, ctx, handlers, evt) {
      var docp, fullPath, obj, objPath, path, value;
      [fullPath, obj, objPath, path, value] = eventPaths(dd, ctx, evt);
      docp = docPath(dd, ctx, fullPath);
      return batch(ctx, function() {
        var name1, ref1, ref2;
        if (typeof handlers[name1 = [obj.type, locationToString(path), "set"].join(',')] === "function") {
          handlers[name1](docp, obj, objPath, path, value, evt);
        }
        return (ref1 = handlers[obj.type]) != null ? (ref2 = ref1[locationToString(path)]) != null ? typeof ref2.set === "function" ? ref2.set(docp, obj, objPath, path, value, evt) : void 0 : void 0 : void 0;
      });
    };
    // patternHandler(DD, CTX, HANDLERS) returns an event handler and makes it easy to define event handlers for types and paths

    // HANDLERS specify event handlers in one of two ways (you can mix them, using whichever is more convenient):
    // - "TYPE,FIELD,EVENT": (OBJ, PATH, KEY, VALUE, EVT)=> ...
    // - TYPE: {FIELD: {EVENT: (OBJ, PATH, KEY, VALUE, EVT)=> ...}}
    patternHandler = function(dd, ctx, handlers) {
      return ctx.handler = {
        clickButton: function(evt) {
          return dispatchClick(dd, ctx, handlers, evt);
        },
        changedValue: function(evt) {
          return dispatchSet(dd, ctx, handlers, evt);
        }
      };
    };
    // #Client Code

    // Connect to WebSocket server
    handleMessage = function(con, [cmd, ...args]) {
      var path, ref1;
      verbose(`Message: ${JSON.stringify([cmd, ...args])}`);
      messages[cmd](con, ...args);
      if (con.batchLevel === 0) {
        ref1 = con.changedJson;
        for (path of ref1) {
          path = stringToLocation(path);
          con.dd.rerender(con.dd.getPath(con.document, con.document.contents, path), Object.assign({}, con.context, {
            location: path
          }), function(dom) {
            if ((dom != null ? dom.getAttribute('data-top') : void 0) != null) {
              return con.dom = dom;
            }
          });
        }
        return con.changedJson.clear();
      }
    };
    connect = function(con, url) {
      var ws;
      con.dd = new Domdom(query('#top'));
      con.batchLevel = 0;
      con.changedJson = new Set();
      con.dom = query('#top');
      con.context = Object.assign({}, con.context, {
        top: null,
        handler: {
          keyPress: function(evt) {
            var key;
            if (key = keyCode(evt)) {
              return ws.send(JSON.stringify(['key', key, stringToLocation(evt.currentTarget.closest('[data-location]').getAttribute('data-location'))]));
            }
          },
          clickButton: function(evt) {
            var name, path;
            path = stringToLocation(evt.currentTarget.getAttribute('data-path-full'));
            name = path.pop();
            return ws.send(JSON.stringify(['click', name, path]));
          },
          changedValue: function(evt, value) {
            var node;
            node = evt.currentTarget;
            return ws.send(JSON.stringify(['set', stringToLocation(node.getAttribute('data-path-full')), value != null ? value : node.value]));
          }
        }
      });
      ws = con.socket = new WebSocket(url);
      ws.onmessage = function(msg) {
        verbose("MESSAGE:", msg);
        return handleMessage(con, JSON.parse(msg.data));
      };
      return ws;
    };
    Object.assign(Domdom, {locationToString, stringToLocation, query, queryAll, find, parseHtml, keyCode, connect, messages, docPath, docPathValue, isDocPath, docPathParts, docPathParent, initChangeContext, batch, change, patternHandler, dispatchClick, dispatchKey, dispatchSet, setVerbose, isRefreshing, takeFocusForScript});
    return Domdom;
  });

}).call(this);

//# sourceMappingURL=domdom.js.map
